/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import de.frankmuenster.mahoe.banks.bcee.Bcee;
import de.frankmuenster.mahoe.banks.bsh.SchwaebischHall;
import de.frankmuenster.mahoe.banks.santander.SantanderBank;
import de.frankmuenster.mahoe.banks.targobank.TargoBank;
import de.frankmuenster.mahoe.jobs.InterfaceMahoeSynchronizeJob;
import de.frankmuenster.mahoe.jobs.MahoeSynchronizeJobProvider;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.web.AbstractSeleniumWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.synchronize.AbstractSynchronizeBackend;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJob;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.jameica.system.OperationCanceledException;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;
import de.willuhn.util.ProgressMonitor;

/**
 * Das Synchronisierungs-Backend der MahoeMain.
 *
 * @author Frank Mnster
 */
public class MahoeSynchronizeBackend extends AbstractSynchronizeBackend<MahoeSynchronizeJobProvider> {

    /**
     * Backend Klassen Name
     */
    public static final String BACKEND_NAME = "de.frankmuenster.mahoe.MahoeSynchronizeBackend"; //$NON-NLS-1$

    /**
     * Der Bank Name, im Moment nur fr Santander Banken
     */
    public static final String BANK_NAME = "Name der Bank"; //$NON-NLS-1$

    /**
     * Die vom Job-Provider explizit untersttzten Bankleitzahlen.
     */
    private static final ConcurrentHashMap<String, String> BLZ_SUPPORTED = new ConcurrentHashMap<>();

    /**
     * The Wallet prefix for this bank.
     */
    private static final ConcurrentHashMap<String, String> WALLET_PREFIXES = new ConcurrentHashMap<>();

    /**
     * VTN fr Schwbisch Hall Vertrge
     */
    public static final String BSH_VTN = "Vertragsnummer (VTN)"; //$NON-NLS-1$

    /**
     * Property name fr das Kreditkartenkonto
     */
    public static final String CARD_ACCOUNT = "Kartenkonto"; //$NON-NLS-1$

    /**
     * Property name fr den Kreditkarten-Anbieter
     */
    public static final String CARD_COMPANY = "Anbieter"; //$NON-NLS-1$

    /**
     * Property name fr die Kreditkartennummer
     */
    public static final String CARD_NUMBER = "Kartennummer"; //$NON-NLS-1$

    /**
     * Konto f&uuml;r die Ausgleichsbuchung
     */
    public static final String CLEARING_ACCOUNT = "Konto-ID der Buchung"; //$NON-NLS-1$

    /**
     * Ausgleichsbuchung anlegen
     */
    public static final String CREATE_CLEARING = "Ausgleichsbuchung ?"; //$NON-NLS-1$

    /**
     * Version des Kontos. Kann fr Updateprozesse verwendet werden.
     */
    public static final String KONTO_VERSION = "KtoVersion"; //$NON-NLS-1$

    /**
     * Max period in the past to load account sheets
     */
    public static final String ACC_STATEM_MAX_DAYS_BACK = // NOPMD
            "Abrufzeitraum Kontoauszuege"; //$NON-NLS-1$ //

    /**
     * Immer nach Passwort fragen
     */
    public static final String PW_ASK_ALWAYS = "Passwort immer nachfragen"; //$NON-NLS-1$

    /**
     * Show the login password just before sending the login request. Only for
     * debug purpose.
     */
    public static final String PW_DEBUG_SHOW = "Passwort anzeigen"; //$NON-NLS-1$

    /**
     * Stay logged in after synchronisation
     */
    public static final String STAY_LOGGED_IN = "Angemeldet bleiben"; //$NON-NLS-1$

    /**
     * Add BLZ to supported BLZ list
     *
     * @param blz
     * @param bankName
     */
    public static void addBlzSupported(final String blz, final String bankName) {

        MahoeSynchronizeBackend.getBlzSupported().putIfAbsent(blz, bankName);

    }

    /**
     * Add Wallet prefix for a bank
     *
     * @param prefix
     * @param bankName
     */
    public static void addWalletPrefix(final String prefix, final String bankName) {

        MahoeSynchronizeBackend.getWalletPrefixs().putIfAbsent(bankName, prefix);

    }

    /**
     * Returns the list with supported BLZs
     *
     * @return the list with supported BLZs
     */
    public static ConcurrentMap<String, String> getBlzSupported() {

        return MahoeSynchronizeBackend.BLZ_SUPPORTED;
    }

    /**
     * Returns the list with Wallet prefixes
     *
     * @return the list with supported BLZs
     */
    public static ConcurrentMap<String, String> getWalletPrefixs() {

        return MahoeSynchronizeBackend.WALLET_PREFIXES;
    }

    /**
     * Standardkonstruktor
     */
    public MahoeSynchronizeBackend() {

        super();
    }

    /**
     * @see de.willuhn.jameica.hbci.synchronize.SynchronizeBackend#getName()
     */
    @Override
    public String getName() {

        return "Mahoe"; //$NON-NLS-1$
    }

    /**
     * @see de.willuhn.jameica.hbci.synchronize.AbstractSynchronizeBackend#getPropertyNames(de.willuhn.jameica.hbci.rmi.Konto)
     */
    @Override
    public List<String> getPropertyNames(final Konto konto) {

        List<String> result = null;
        try {
            if (konto != null && !konto.hasFlag(Konto.FLAG_DISABLED)) {
                if (SantanderBank.BLZ.equals(konto.getBLZ())) {
                    result = new ArrayList<>();
                    result.add(MahoeSynchronizeBackend.BANK_NAME);
                    // Mahoe-149: further properties moved to settings
                } else if (TargoBank.BLZ.equals(konto.getBLZ())) {
                    result = new ArrayList<>();
                    result.add(MahoeSynchronizeBackend.CARD_ACCOUNT);
                    result.add(MahoeSynchronizeBackend.CARD_NUMBER);
                    result.add(MahoeSynchronizeBackend.CARD_COMPANY);
                } else if (SchwaebischHall.BLZ.equals(konto.getBLZ())) {
                    result = new ArrayList<>();
                    result.add(MahoeSynchronizeBackend.BSH_VTN);
                } else if (Bcee.BIC.equals(konto.getBic())) { // NOPMD
                }
            }
        } catch (final RemoteException re) {
            Logger.error(I18N.tr("MahoeSynchronizeBackend.propertyNameNotFound"), re); //$NON-NLS-1$
        }
        return result;
    }

    /**
     * @see de.willuhn.jameica.hbci.synchronize.AbstractSynchronizeBackend#supports(java.lang.Class,
     *      de.willuhn.jameica.hbci.rmi.Konto)
     */
    @Override
    public boolean supports(final Class<? extends SynchronizeJob> type, final Konto konto) {

        boolean support = false;

        if (super.supports(type, konto)) {
            try {
                Logger.debug("Support BLZ: " + konto.getBLZ()); //$NON-NLS-1$
                Logger.debug(I18N.tr("MahoeSynchronizeBackend.jobType") + type.getName()); //$NON-NLS-1$
                // Checken, ob wir die Bank untersttzen
                if (MahoeSynchronizeBackend.getBlzSupported().containsKey(konto.getBLZ())) {
                    support = true;
                }
            } catch (final RemoteException re) {
                Logger.error(I18N.tr("MahoeSynchronizeBackend.checkSupportedFailed"), re); //$NON-NLS-1$
            }
        }

        return support;
    }

    /**
     * @see de.willuhn.jameica.hbci.synchronize.AbstractSynchronizeBackend#createJobGroup(de.willuhn.jameica.hbci.rmi.Konto)
     */
    @Override
    protected AbstractSynchronizeBackend<MahoeSynchronizeJobProvider>.JobGroup createJobGroup(final Konto konto) {

        return new MahoeJobGroup(konto);
    }

    /**
     * @see de.willuhn.jameica.hbci.synchronize.AbstractSynchronizeBackend#getJobProviderInterface()
     */
    @Override
    protected Class<MahoeSynchronizeJobProvider> getJobProviderInterface() {

        return MahoeSynchronizeJobProvider.class;
    }

    /**
     * Hier ist die Implementierung unseres Jobs
     *
     * @author Frank Mnster
     */
    @SuppressWarnings({ "PMD.StdCyclomaticComplexity", "PMD.CyclomaticComplexity" })
    protected class MahoeJobGroup extends JobGroup {

        /**
         * @param konto
         *            Das Konto
         */
        protected MahoeJobGroup(final Konto konto) {

            super(konto);
        }

        /**
         * @throws WebSiteException
         * @see de.willuhn.jameica.hbci.synchronize.AbstractSynchronizeBackend.JobGroup#sync()
         */
        @Override
        protected void sync() throws Exception { // NOPMD

            final ProgressMonitor monitor = MahoeSynchronizeBackend.this.worker.getMonitor();
            final String ktoLang = this.getKonto().getLongName();
            final int step = 100 / MahoeSynchronizeBackend.this.worker.getSynchronization().size();
            monitor.log(" "); //$NON-NLS-1$
            monitor.log(I18N.tr("MahoeSynchronizeBackend.syncAccountMsg", ktoLang)); //$NON-NLS-1$
            Logger.info(I18N.tr("MahoeSynchronizeBackend.syncAccountMsg", ktoLang)); //$NON-NLS-1$
            int jobsCount = 0;
            // Java, OS, HTMLUnit and Mahoe versions
            final String text = "Java: " //$NON-NLS-1$
                    + System.getProperty("java.version") + " | OS: " //$NON-NLS-1$ //$NON-NLS-2$
                    + System.getProperty("os.name") + " " //$NON-NLS-1$ //$NON-NLS-2$
                    + System.getProperty("os.arch") + " | HTMLUnit: " //$NON-NLS-1$ //$NON-NLS-2$
                    + com.gargoylesoftware.htmlunit.Version.getProductVersion() + " | System-Encoding : " //$NON-NLS-1$
                    + System.getProperty("file.encoding") + " | Mahoe: " //$NON-NLS-1$ //$NON-NLS-2$
                    + Application.getPluginLoader().getPlugin(MahoeMain.class).getManifest().getVersion() + "(" //$NON-NLS-1$
                    + Application.getPluginLoader().getPlugin(MahoeMain.class).getManifest().getBuildnumber() + ")"; //$NON-NLS-1$
            Logger.info(text);
            monitor.log(text);
            for (final SynchronizeJob job : this.jobs) {
                this.checkInterrupted();
                jobsCount++;
                try {
                    ((InterfaceMahoeSynchronizeJob) job).execute(monitor, jobsCount == this.jobs.size());
                } catch (final OperationCanceledException e) {
                    Logger.debug(e.getMessage());
                    throw e;
                } catch (final WebSiteException e) {
                    switch (e.getType()) {
                    case LOGIN_FAILED:
                        String msg = I18N.tr("MahoeSynchronizeBackend.loginFailed"); //$NON-NLS-1$
                        Logger.error(msg, e);
                        Application.getCallback().notifyUser(msg);
                        StatusBarMessage msgUser = new StatusBarMessage(msg, StatusBarMessage.TYPE_INFO); // NOPMD
                        Application.getMessagingFactory().sendMessage(msgUser);
                        monitor.setStatusText(msg);
                        throw e;
                    case LOGIN_ERROR:
                        msg = I18N.tr("MahoeSynchronizeBackend.loginError"); //$NON-NLS-1$
                        Logger.error(msg, e);
                        Application.getCallback().notifyUser(msg);
                        monitor.setStatusText(msg);
                        throw e;
                    case SERVICE_UNAVAILABLE:
                        msg = I18N.tr("MahoeSynchronizeBackend.serviceUnavailable"); //$NON-NLS-1$
                        Logger.error(msg, e);
                        Application.getCallback().notifyUser(msg);
                        msgUser = new StatusBarMessage(msg, // NOPMD
                                StatusBarMessage.TYPE_ERROR);
                        Application.getMessagingFactory().sendMessage(msgUser);
                        monitor.setStatusText(I18N.tr("MahoeSynchronizeBackend.serviceUnavailable")); //$NON-NLS-1$
                        monitor.setStatus(ProgressMonitor.STATUS_ERROR);
                        throw e;
                    case TRANSFER_FAILED:
                        msg = e.getMessage(); // $NON-NLS-1$
                        Logger.error(msg, e);
                        Application.getCallback().notifyUser(msg);
                        msgUser = new StatusBarMessage(msg, // NOPMD
                                StatusBarMessage.TYPE_ERROR); // NOPMD
                        Application.getMessagingFactory().sendMessage(msgUser);
                        monitor.setStatusText("berweisung konnte nicht eingereicht werden."); //$NON-NLS-1$
                        monitor.setStatus(ProgressMonitor.STATUS_ERROR);
                        throw e;
                    default:
                        Logger.error(I18N.tr("MahoeSynchronizeBackend.getTransactionsFailedLog", //$NON-NLS-1$
                                e.getMessage()), e);
                        msg = I18N.tr("MahoeSynchronizeBackend.getTransactionsFailedMsg"); //$NON-NLS-1$
                        msgUser = new StatusBarMessage(msg, // NOPMD
                                StatusBarMessage.TYPE_ERROR); // NOPMD
                        Application.getMessagingFactory().sendMessage(msgUser);
                        throw e;
                    }
                } catch (final ApplicationException e) {
                    Logger.error(e.getMessage(), e);
                    throw e;
                } catch (final Exception e) { // NOPMD
                    Logger.error(I18N.tr("MahoeSynchronizeBackend.generalExpectionOccured"), e); //$NON-NLS-1$
                    throw new ApplicationException(I18N.tr("MahoeSynchronizeBackend.generalExpectionOccured"), e); //$NON-NLS-1$
                }
                monitor.addPercentComplete(step);
            }

            // Cleanup Selenium
            if (Settings.isSeleniumAutoClose()) {
                AbstractSeleniumWebSite.shutDown();
            }
        }
    }
}
