/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.bsh.assistant;

import java.rmi.RemoteException;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.assistant.AbstractCreateAccountAssistant;
import de.frankmuenster.mahoe.assistant.CreateAccountException;
import de.frankmuenster.mahoe.banks.bsh.SchwaebischHall;
import de.frankmuenster.mahoe.banks.bsh.json.Account;
import de.frankmuenster.mahoe.banks.bsh.web.BshWebSite;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.system.OperationCanceledException;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * BSH Konto-Anlagen-Assistent
 *
 * @author Frank Mnster
 */
public final class BshCreateAccountAssistent
        extends AbstractCreateAccountAssistant<BshCreateInfo, BshCreateAccountDialog> {

    /**
     * Standardkonstruktor
     */
    public BshCreateAccountAssistent() {

        super();
    }

    /**
     * Loads all existing accounts for a given BSH login and populates this in the <code>BshCreateAccountDialog</code>
     *
     * @see AbstractCreateAccountAssistant#getDialog()
     */
    @Override
    public BshCreateAccountDialog getDialog() throws CreateAccountException {

        final BshCreateKontoStartDialog startDialog = new BshCreateKontoStartDialog();
        String kdNummer;
        try {
            kdNummer = startDialog.open();
        } catch (final OperationCanceledException e) { // NOPMD
            throw e;
        } catch (final Exception e) { // NOPMD
            Logger.warn(I18N.tr("BshCreateAccountAssistent.openDialogFailedLog")); //$NON-NLS-1$
            throw new CreateAccountException(I18N.tr("BshCreateAccountAssistent.openDialogFailed"), e); //$NON-NLS-1$
        }
        if (StringUtils.isEmpty(kdNummer)) {
            throw new CreateAccountException(I18N.tr("BshCreateAccountAssistent.kdNumberNotProvided")); //$NON-NLS-1$
        }

        List<Account> accounts = null;
        try {
            accounts = this.getAccounts(kdNummer);
        } catch (WebSiteException | ApplicationException e1) {
            Logger.error(e1.getMessage(), e1);
            throw new CreateAccountException(e1.getMessage(), e1);
        }

        return new BshCreateAccountDialog(kdNummer, accounts);
    }

    /**
     * @see AbstractCreateAccountAssistant#setBankSpecificMetaData(de.frankmuenster.mahoe.assistant.CreateInfo,
     *      Konto)
     */
    @Override
    public void setBankSpecificMetaData(final BshCreateInfo info, final Konto konto)
            throws RemoteException, ApplicationException {

        konto.setMeta(MahoeSynchronizeBackend.BSH_VTN, info.getVtn());

    }

    @Override
    protected void setBankSpecificAccountParameter(final Konto konto, final BshCreateInfo info) throws RemoteException {

        // account settings
        konto.setBLZ(SchwaebischHall.BLZ);
        konto.setIban(this.getIban(info));
        konto.setKontonummer(info.getAccount().getCoreId());

        konto.setKundennummer(info.getKdNummer());
        konto.setUnterkonto(info.getLetterNumber());

    }

    /**
     * Ermittelt die Liste aller Vertrge der Kundennummer
     *
     * @param kdNummer
     * @return die Liste aller Vertrge
     * @throws ApplicationException
     * @throws WebSiteException
     */
    private List<Account> getAccounts(final String kdNummer) throws WebSiteException, ApplicationException {

        final BshWebSite webSite = BshWebSite.getInstance(kdNummer);
        return webSite.getAccounts();
    }

    /**
     * Ermittelt die IBAN.
     *
     * @param info
     * @return die IBAN, oder NULL
     */
    private String getIban(final BshCreateInfo info) {

        String iban = null;
        try {
            final BshWebSite webSite = BshWebSite.getInstance(info.getKdNummer());
            final String bsnKurz = info.getCoreId() + info.getLetterNumber();
            iban = webSite.getIban(bsnKurz, info.getBlzDz());
            if (iban != null) {
                // Evtl. Leerzeichen entfernen
                iban = iban.replaceAll(" ", StringUtils.EMPTY); //$NON-NLS-1$
            }
        } catch (WebSiteException | ApplicationException e) {
            // Nur Loggen
            Logger.error(e.getMessage(), e);
        }

        return iban;
    }
}
