/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.web;

import java.rmi.RemoteException;
import java.util.Date;

import org.apache.commons.lang3.tuple.Pair;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;

import de.frankmuenster.mahoe.banks.santander.SantanderBank;
import de.frankmuenster.mahoe.banks.santander.web.SantanderWebSite.DocumentType;
import de.frankmuenster.mahoe.banks.santander.webpages.PostBoxPage;
import de.frankmuenster.mahoe.statements.AbstractAccountStatementFinder;
import de.frankmuenster.mahoe.statements.MahoeAccountStatement;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.KontoType;
import de.willuhn.logging.Logger;

/**
 * Class to get account statements from the banks web page
 *
 * @author Frank Mnster
 */
@SuppressWarnings("PMD.ExcessiveImports")
public class SantanderAccountStatementFinder
        extends AbstractAccountStatementFinder<WebDriver, SantanderWebAccountInformation> {

    /**
     * Document type to download
     */
    private final DocumentType docType;

    /**
     * The baseUrl for all requests in this session
     */
    private final String baseUrl;

    /**
     * Creates a new SantanderAccountStatement object with a webClient and
     * account information
     *
     * @param webClient
     * @param accountInfo
     * @param baseUrl
     * @throws WebSiteException
     */
    public SantanderAccountStatementFinder(final WebDriver webClient, final String baseUrl,
            final SantanderWebAccountInformation accountInfo) throws WebSiteException {

        super(webClient, accountInfo);
        this.baseUrl = baseUrl;
        this.docType = this.getDocType();

    }

    /**
     * @see AbstractAccountStatementFinder#startSearch()
     */
    @Override
    protected void startSearch() throws WebSiteException {

        try {

            // open post box
            final PostBoxPage postBox = new PostBoxPage(this.webClient, this.baseUrl, this.accountInfo);

            // loop over entries
            while (postBox.hasNextKtoCard()) {

                // get the next document link
                final Pair<Date, WebElement> document = postBox.nextNextKtoCard();

                // check date
                if (document.getKey().before(this.lastStatementDate)) {
                    Logger.debug(I18N.tr("TargoBankWebSiteTransactionsImpl.reachedLastPdf")); //$NON-NLS-1$
                    break;
                }

                // get the PDF as byte array
                final byte[] pdf = SantanderWebSite.getAccountSheetAsByteArray(document, this.webClient);

                // create a mahoe statement object
                final MahoeAccountStatement statementInfo = new MahoeAccountStatement(document.getKey(), // NOPMD
                        SantanderBank.NAME, this.docType.getValue());
                // add to list
                this.addStatement(statementInfo, pdf);
            }
        } catch (final RemoteException e) {
            Logger.error(I18N.tr("AccountStatementFinder.downloadFailed"), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("AccountStatementFinder.downloadFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.WEBSITE_NOT_REACHABLE, e);
        }

    }

    /**
     * Returns the document type to download based on account type
     *
     * @return the {@link DocumentType} to download
     * @throws WebSiteException
     */
    private DocumentType getDocType() throws WebSiteException {

        DocumentType dType = null;
        try {
            if (KontoType.GIRO.getValue() == this.accountInfo.getKonto().getAccountType()) { // NOPMD
                dType = DocumentType.KONTOAUSZUG;
            } else if (KontoType.KREDITKARTE.getValue() == this.accountInfo.getKonto().getAccountType()) {
                dType = DocumentType.KARTENABRECHNUNG;
            } else {
                throw new WebSiteException(I18N.tr("AccountStatementFinder.accountTypeNotSupported"), //$NON-NLS-1$
                        WebSiteExceptionTypes.ACCOUNT_INFO_WRONG);
            }
        } catch (final RemoteException e) {
            throw new WebSiteException(I18N.tr("AccountStatementFinder.failedGettingAccountNumber"), e); //$NON-NLS-1$
        }
        return dType;
    }
}
