/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.webpages;

import java.util.List;

import org.openqa.selenium.By;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebDriverException;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import de.frankmuenster.mahoe.banks.santander.SantanderBank;
import de.frankmuenster.mahoe.banks.santander.web.SantanderWebSite;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.SecureCharSequence;
import de.frankmuenster.mahoe.utils.SeleniumCookie;
import de.frankmuenster.mahoe.utils.WebUtilsSelenium;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.logging.Logger;

/**
 * The login page for the Santander Bank
 *
 * @author Frank Mnster
 * @since 2.4.0
 *
 */
public class LoginPage {

    /**
     * Selenium web driver
     */
    private final WebDriver driver;

    /**
     * Check if a user is logged in. There is no check if a specific account is
     * logged in.
     *
     * @return <code>true</code> if an user is logged in, <code>false</code>
     *         else.
     */
    public static boolean isLoggedIn(final WebDriver driver) {

        boolean isLoggedIn = false;
        try {

            // first test if we are on the correct bank
            final String url = driver.getCurrentUrl();
            if (url.contains("santander.de")) { //$NON-NLS-1$
                final String check = driver.getPageSource();
                if (check.contains("Abmelden")) { //$NON-NLS-1$
                    isLoggedIn = true;
                    // Fall back to home page
                    driver.get(SantanderBank.LOGIN_URL + "/#/"); //$NON-NLS-1$
                    // wait for loading is finished
                    SantanderWebSite.waitForLoadingOverlay(driver);
                }
            }
        } catch (final WebDriverException e) {
            // Only log
            Logger.debug("Selenium session already quited." + e.getLocalizedMessage()); //$NON-NLS-1$
        }

        return isLoggedIn;

    }

    /**
     * Creates a new LoginPage for the Santander Bank
     *
     * @param driver
     *            the {@link WebDriver} to use for that page
     */
    public LoginPage(final WebDriver driver) {
        this.driver = driver;
    }

    /**
     * Logging the given user.
     *
     * @param login
     *            the account to login
     * @param password
     *            the password
     * @throws WebSiteException
     *             if the login fails
     */
    public void login(final String login, final char[] password) throws WebSiteException { // NOPMD

        // check if not already logged-in
        if (!LoginPage.isLoggedIn(this.driver)) {

            // open login page
            final WebDriverWait waitLogin = new WebDriverWait(this.driver, 120);
            this.driver.get(SantanderBank.LOGIN_URL);
            waitLogin.until(ExpectedConditions.visibilityOfElementLocated(By.id("alias"))); //$NON-NLS-1$

            // check for cookie acceptance button
            checkCookieWarning();

            // reload cookies
            SeleniumCookie.read(this.driver, SantanderBank.COOKIE_FILE);

            // search for the submit button
            final WebElement submitBtn = getSubmmitButton();

            // is submit button found
            if (submitBtn == null) {
                Logger.error(I18N.tr("LoginPage.loginBtnNotFound")); //$NON-NLS-1$
                throw new WebSiteException(I18N.tr("LoginPage.loginBtnNotFound"), WebSiteExceptionTypes.LOGIN_ERROR); //$NON-NLS-1$
            }

            // get fields for alias and password
            final WebElement alias = this.driver.findElement(By.id("alias")); //$NON-NLS-1$
            final WebElement passwd = this.driver.findElement(By.id("password")); //$NON-NLS-1$

            // Set alias
            Logger.debug("Setzte Kundenummer auf: ****" + login.substring(login.length() - 4)); //$NON-NLS-1$
            alias.sendKeys(login);
            // Set password
            final SecureCharSequence securePW = new SecureCharSequence(password);
            passwd.sendKeys(securePW);
            securePW.wipe();

            // Scroll and click on submit login form
            WebUtilsSelenium.scrollTo(this.driver, submitBtn);
            submitBtn.click();

            // wait for the overlay to be removed
            SantanderWebSite.waitForLoadingOverlay(this.driver);

            // wait until the logout button is found
            waitLogin.until(ExpectedConditions.or(
                    ExpectedConditions.presenceOfElementLocated(By.className("loggedInHeader__logoutButton___1sduq")), //$NON-NLS-1$
                    ExpectedConditions.presenceOfElementLocated(By.className("formMessage__formMessage___1nrJ8")))); //$NON-NLS-1$

            // do a final check, if login was successful
            final String check = this.driver.getPageSource();

            if (check.contains("Ihr Login war nicht erfolgreich.")) { //$NON-NLS-1$
                throw new WebSiteException(I18N.tr("Login.wrongCredentials"), WebSiteExceptionTypes.LOGIN_FAILED); //$NON-NLS-1$
            }
            if (!check.contains("Finanzbersicht")) { //$NON-NLS-1$
                throw new WebSiteException(I18N.tr("TargoBankWebSite.loginFailed"), WebSiteExceptionTypes.LOGIN_ERROR); //$NON-NLS-1$
            }

        }
    }

    /**
     * Check if the cookie warning is displayed.If yes the acceptance button is
     * clicked.
     *
     * @param driver
     *            the {@link WebDriver} with the page (login?) to check for
     *            cookie warning
     */
    private void checkCookieWarning() {
        final List<WebElement> cookieBtn = this.driver.findElements(By.className("cookieLayer__buttonWrapper___q5ioB")); //$NON-NLS-1$
        if (!cookieBtn.isEmpty()) {
            // Cookie banner is visible -- find button
            final WebElement btn = cookieBtn.get(0).findElement(By.tagName("button")); //$NON-NLS-1$
            if (btn != null) {
                // acceptance button found, now click
                btn.click();
            }
        }
    }

    /**
     * Search for a span-element with the text "Einloggen".
     *
     * @param driver
     *            the {@link WebDriver} with a page loaded to search for the
     *            element
     * @return the element if found, <i>null</i> else
     */
    private WebElement getSubmmitButton() {

        // Initialise the element with NULL
        WebElement submitBtn = null;
        // get all span-elements
        final List<WebElement> spans = this.driver.findElements(By.tagName("span")); //$NON-NLS-1$
        // Loop over all elements
        for (final WebElement webElement : spans) {
            // extract the text from the element
            final String text = webElement.getText().trim();
            Logger.debug(text);
            // Check if text is "Einloggen"
            if ("Einloggen".equalsIgnoreCase(text)) { //$NON-NLS-1$
                // assign to returning element
                submitBtn = webElement;
                // leave the loop as we have found teh element
                break;
            }
        }
        // return the element
        return submitBtn;
    }

}
