/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.webpages;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.NoSuchElementException;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.openqa.selenium.By;
import org.openqa.selenium.ElementNotVisibleException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import com.gargoylesoftware.htmlunit.ElementNotFoundException;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.banks.santander.web.SantanderMonthEnum;
import de.frankmuenster.mahoe.banks.santander.web.SantanderWebAccountInformation;
import de.frankmuenster.mahoe.banks.santander.web.SantanderWebSite;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.KontoType;
import de.willuhn.logging.Logger;

/**
 * The post box page
 *
 * @author Frank Mnster
 * @since 2.4.0
 *
 */
public class PostBoxPage {

    /**
     * PostBox URL
     */
    private static final String POST_BOX_URL_SUFFIX = "/postbox"; //$NON-NLS-1$

    /**
     * The webdriver with logged-in user
     */
    protected final WebDriver driver;

    /**
     * Account to open the details page for
     */
    protected final SantanderWebAccountInformation accountInfo;

    /**
     * Account type
     */
    protected final KontoType ktoType;

    /**
     * Base URL for all following requests
     */
    private final String baseUrl;

    /**
     * Current position in ktoCard list
     */
    private int ktoCardPosition;

    /**
     * Current position in other documents list
     */
    private int documentsPosition;

    /**
     * The actual year
     */
    private int year;

    /**
     * The month of last transaction. Used to calculate the year
     */
    private int lastDocumentMonth;

    /**
     * The month of last transaction. Used to calculate the year
     */
    private int lastDocumentDay;

    /** List with all account sheets/card billings */
    private List<Pair<Date, WebElement>> ktoCardDocuments;

    /** List with all other documents */
    private List<Pair<Date, WebElement>> otherDocuments;

    /**
     * Create a new post box page for the given account
     *
     * @param driver
     * @param accountInfo
     * @param baseUrl
     *
     * @throws WebSiteException
     * @throws RemoteException
     */
    public PostBoxPage(final WebDriver driver, final String baseUrl, final SantanderWebAccountInformation accountInfo)
            throws WebSiteException, RemoteException {

        this.driver = driver;
        this.accountInfo = accountInfo;
        this.baseUrl = baseUrl;
        this.ktoType = KontoType.find(accountInfo.getKonto().getAccountType());

        String ktoCardNumber;
        Identifier identifier;
        // get information for accounts/ credit cards
        if (KontoType.KREDITKARTE.equals(this.ktoType)) {
            ktoCardNumber = accountInfo.getKonto().getMeta(MahoeSynchronizeBackend.CARD_NUMBER, StringUtils.EMPTY);
            identifier = Identifier.KREDITKARTE;
        } else {
            ktoCardNumber = accountInfo.getKonto().getIban();
            identifier = Identifier.KONTO;
        }

        // open Postbox
        openPage();
        // Load all documents in a list
        loadDocuments(ktoCardNumber, identifier);

    }

    /**
     * Calculates the date based on the given month and date
     *
     * @param month
     * @param date
     * @return the date
     */

    private Date calculateDate(final String month, final String date) throws RemoteException {

        // create new calendar
        final Calendar cal = Calendar.getInstance();

        // set the day
        final int transactionDay = Integer.parseInt(date);
        cal.set(Calendar.DATE, transactionDay);

        // calculate month and set the month
        final int transactionMonth = SantanderMonthEnum.getNumber(month);
        if (transactionMonth == -1) {
            throw new RemoteException(I18N.tr("PostBoxPage.monthOutOfRange")); //$NON-NLS-1$
        } else {
            cal.set(Calendar.MONTH, transactionMonth);
        }

        // check if year to be changed
        if (transactionMonth > this.lastDocumentMonth
                || transactionMonth == this.lastDocumentMonth && transactionDay >= this.lastDocumentDay) {
            this.year--;
        }
        // set the year
        cal.set(Calendar.YEAR, this.year);

        // update last transaction month and day
        this.lastDocumentMonth = transactionMonth;
        this.lastDocumentDay = transactionDay;

        // return as date
        return cal.getTime();
    }

    /**
     * Returns {@code true} if the PostBox has more accounts sheets/card
     * billings. (In other words, returns {@code true} if {@link #nextDocument}
     * would return an element rather than throwing an exception.)
     *
     * @return {@code true} if the PostBox has more elements
     */
    public boolean hasNextDocument() {
        return this.documentsPosition < this.otherDocuments.size();
    }

    /**
     * Returns {@code true} if the PostBox has more accounts sheets/card
     * billings. (In other words, returns {@code true} if {@link #nextDocument}
     * would return an element rather than throwing an exception.)
     *
     * @return {@code true} if the PostBox has more elements
     */
    public boolean hasNextKtoCard() {
        return this.ktoCardPosition < this.ktoCardDocuments.size();
    }

    private void loadDocuments(final String ktoCardNumber, final Identifier identifier) throws RemoteException {

        this.ktoCardDocuments = new ArrayList<>();
        this.otherDocuments = new ArrayList<>();

        final List<
                WebElement> docElements = this.driver.findElements(By.className("postboxDateWrapper__wrapper___1kfAU")); //$NON-NLS-1$

        // Initialise list positions
        this.ktoCardPosition = 0;
        this.documentsPosition = 0;

        // set actual year, month and day for year calculations
        this.year = Calendar.getInstance().get(Calendar.YEAR);
        this.lastDocumentMonth = Calendar.getInstance().get(Calendar.MONTH);
        this.lastDocumentDay = Calendar.getInstance().get(Calendar.DAY_OF_MONTH) + 1;

        // Loop over documents
        for (final WebElement listElement : docElements) {

            // get documents text to extract date and card number
            final String test = listElement.getText();
            Logger.debug("Dokument als Text: " + test); //$NON-NLS-1$

            // spilt in lines
            final String[] parts = test.split("\n"); //$NON-NLS-1$

            // kto/card number is part 3
            String ktoCard = parts[3];
            if (identifier == Identifier.KREDITKARTE) {
                ktoCard = ktoCard.substring(ktoCard.length() - 4, ktoCard.length());
            } else {
                ktoCard = StringUtils.deleteWhitespace(ktoCard);
            }

            // continue only if kto/card number matches
            if (ktoCardNumber.endsWith(ktoCard)) {

                // set date and valuta to the same as we have only one date
                final Date date = calculateDate(parts[0], parts[1]);

                // create the pair and add document to list
                final Pair<Date, WebElement> document = Pair.of(date, listElement);
                this.ktoCardDocuments.add(document);
            } else if ("Wichtige Informationen".equals(parts[2])) { //$NON-NLS-1$
                // set date and valuta to the same as we have only one date
                final Date date = calculateDate(parts[0], parts[1]);

                // create the pair and add document to list
                final Pair<Date, WebElement> document = Pair.of(date, listElement);
                this.otherDocuments.add(document);
            }
        }
    }

    /**
     * Returns the next document from the PostBox
     *
     * @return the next document as Pair< Date,WebElement >
     * @throws NoSuchElementException
     *             - if the PostBox has no more elements
     *
     */
    public Pair<Date, WebElement> nextDocument() throws NoSuchElementException {

        this.documentsPosition++;
        if (this.documentsPosition > this.otherDocuments.size()) {
            throw new NoSuchElementException(I18N.tr("PostBoxPage.noMoreDocuments")); //$NON-NLS-1$
        }

        return this.otherDocuments.get(this.documentsPosition - 1);

    }

    /**
     * Returns the next accounts sheet/card billing from the PostBox
     *
     * @return the next accounts sheets/card billings as Pair< Date,WebElement >
     * @throws NoSuchElementException
     *             - if the PostBox has no more elements
     *
     */
    public Pair<Date, WebElement> nextNextKtoCard() throws NoSuchElementException {

        this.ktoCardPosition++;
        if (this.ktoCardPosition > this.ktoCardDocuments.size()) {
            throw new NoSuchElementException(I18N.tr("PostBoxPage.noMoreDocuments")); //$NON-NLS-1$
        }

        return this.ktoCardDocuments.get(this.ktoCardPosition - 1);

    }

    /**
     * Open the page for the given account.
     *
     * @param ktoCardNumber
     *
     * @throws WebSiteException
     *             in case the page cannot be opened or the account is not found
     */
    private void openPage() throws WebSiteException {

        try {
            final String callUrl = this.baseUrl + PostBoxPage.POST_BOX_URL_SUFFIX; // $NON-NLS-1$

            this.driver.get(callUrl);

            final WebDriverWait wait10Sec = new WebDriverWait(this.driver, 120);

            SantanderWebSite.waitForLoadingOverlay(this.driver);

            wait10Sec
                    .until(ExpectedConditions.presenceOfElementLocated(By.className("postboxHeader__wrapper___2Z-IH"))); //$NON-NLS-1$ )

            Logger.debug(this.driver.getPageSource());

        } catch (final ElementNotFoundException | ElementNotVisibleException e) {
            Logger.error(I18N.tr("PostBoxPage.errorOpeningPostBox"), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("PostBoxPage.errorOpeningPostBox"), //$NON-NLS-1$
                    WebSiteExceptionTypes.PAGE_NOT_FOUND, e);
        }

    }

    /**
     * Identifier for
     */
    private enum Identifier {
        KONTO("Kontodokumente"), //$NON-NLS-1$
        KREDITKARTE("Kreditkartendokumente"); //$NON-NLS-1$

        /** Text on web page */
        private String value;

        Identifier(final String value) {
            this.value = value;
        }

        @SuppressWarnings("unused")
        public String getValue() {
            return this.value;
        }
    }

}
