/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.webpages;

import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.openqa.selenium.By;
import org.openqa.selenium.ElementNotVisibleException;
import org.openqa.selenium.TimeoutException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import com.gargoylesoftware.htmlunit.ElementNotFoundException;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.banks.santander.web.SantanderMonthEnum;
import de.frankmuenster.mahoe.banks.santander.web.SantanderWebAccountInformation;
import de.frankmuenster.mahoe.banks.santander.web.SantanderWebSite;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.utils.WebUtilsSelenium;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.KontoType;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;

/**
 * The page with all listed transactions These are the booked transactions for
 * giro account and the pre-booked for an credit card account.
 *
 *
 * @author Frank Mnster
 * @since 2.4.0
 *
 */
@SuppressWarnings("PMD.ExcessiveImports")
public class TransactionsPage {

    /**
     * "Datum" sting to be defiend only once
     */
    private static final String DATUM_STRING = "Datum:"; //$NON-NLS-1$

    /** Debug text to log an transaction text completely */
    private static final String LOG_UMSATZ_ALS_TEXT = "Umsatz als Text: "; //$NON-NLS-1$

    /**
     * The webdriver with logged-in user
     */
    protected final WebDriver driver;

    /**
     * Account to open the details page for
     */
    protected final SantanderWebAccountInformation accountInfo;

    /**
     * Account type
     */
    protected final boolean isCreditCard;

    /**
     * Base URL for all following requests
     */
    private final String baseUrl;

    /**
     * The actual year
     */
    private int year;

    /**
     * The month of last transaction. Used to calculate the year
     */
    private int lastTransactionMonth;

    /**
     * Create a new account page for the given account
     *
     * @param driver
     * @param accountInfo
     * @param baseUrl
     *
     * @throws WebSiteException
     * @throws RemoteException
     */
    public TransactionsPage(final WebDriver driver, final String baseUrl,
            final SantanderWebAccountInformation accountInfo) throws WebSiteException {

        this.driver = driver;
        this.accountInfo = accountInfo;
        this.baseUrl = baseUrl;

        try {
            final KontoType ktoType = KontoType.find(accountInfo.getKonto().getAccountType());

            // try to open the page
            if (KontoType.KREDITKARTE.equals(ktoType)) {
                final String cardNumber = accountInfo.getKonto().getMeta(MahoeSynchronizeBackend.CARD_NUMBER,
                        StringUtils.EMPTY);
                openPage(cardNumber);
                this.isCreditCard = true;
            } else {
                openPage(accountInfo.getIban());
                this.isCreditCard = false;
            }
        } catch (final RemoteException e) {
            throw new WebSiteException(I18N.tr("AccountPage.errorGetingAccountDetails"), //$NON-NLS-1$
                    WebSiteExceptionTypes.ACCOUNT_INFO_WRONG, e);
        }

    }

    /**
     * Returns all transactions on current page till the given date
     *
     * @param Date
     *            lastTransactionDate the of the last transactions to give back
     * @throws WebSiteException
     */
    public List<Umsatz> getTransactions(final Date lastTransactionDate) throws WebSiteException {

        List<Umsatz> transactions;
        try {
            // get list with transactions
            WebElement list;
            if (this.isCreditCard) {
                list = this.driver.findElement(By.className("globalPosition__accountDetailWp___1Gx4Z")); //$NON-NLS-1$
            } else {
                list = this.driver.findElement(By.className("accountsListItem__listContainer___11bYm")); //$NON-NLS-1$
            }

            // expand list as it lazy loading with ajax
            final List<WebElement> listElements = expandList(list);

            // get the date-web element pairs
            final List<Pair<Date, WebElement>> transactionList = getTransactionList(listElements);

            // add all transactions
            transactions = getTransactionsFromPairList(transactionList, lastTransactionDate);

        } catch (final RemoteException e) {
            Logger.error(I18N.tr("TransactionPage.getPreBookedOverviewFailedLog", //$NON-NLS-1$
                    e.getMessage()), e);
            throw new WebSiteException(I18N.tr("TransactionPage.getPreBookedOverviewFailed"), // NOPMD //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return transactions;
    }

    /**
     * Calculates the date based on the given month and date
     *
     * @param month
     * @param date
     * @return the date
     */

    private Date calculateDate(final String month, final String date) throws RemoteException {

        // create new calendar
        final Calendar cal = Calendar.getInstance();

        // set the day
        cal.set(Calendar.DATE, Integer.valueOf(date));

        // calculate month and set the month
        final int transactionMonth = SantanderMonthEnum.getNumber(month);
        if (transactionMonth == -1) {
            throw new RemoteException(I18N.tr("AccountPage.monthOutOfRange")); //$NON-NLS-1$
        }
        cal.set(Calendar.MONTH, transactionMonth);

        // check if year to be changed
        if (transactionMonth > this.lastTransactionMonth) {
            this.year--;
        }
        // set the year
        cal.set(Calendar.YEAR, this.year);

        // update last transaction month
        this.lastTransactionMonth = transactionMonth;

        // return as date
        return cal.getTime();
    }

    /**
     * Expand the movements list until no further notice is displayed any more
     *
     * @param list
     *            initial list
     * @return a list with all movements from the page
     */
    private List<WebElement> expandList(final WebElement list) {

        // does load more button exist
        List<WebElement> buttons = this.driver.findElements(By.className("momentsOverview__loadMoreBtn___2gLGP")); //$NON-NLS-1$

        // create a new waiter
        final WebDriverWait wait = new WebDriverWait(this.driver, 2);

        while (!buttons.isEmpty()) {

            // get the button
            final WebElement button = buttons.get(0);

            // button exists, scroll into view
            WebUtilsSelenium.scrollTo(this.driver, button);
            // wait for any overlay
            SantanderWebSite.waitForLoadingOverlay(this.driver);

            // Wait until link is visibile
            try {
                // wait max. two seconds
                wait.until(ExpectedConditions.visibilityOf(button));
            } catch (final TimeoutException e) {
                Logger.debug(String.format("Weiter-Button nicht sichtbar. Fehler: %1$s", e.getLocalizedMessage())); //$NON-NLS-1$
                // leave he loop in case the button is not clickable
                break;
            }

            // click the button to load further movements
            button.click();
            // wait again for the oevrlay
            SantanderWebSite.waitForLoadingOverlay(this.driver);
            // re-check if the button still exists
            buttons = list.findElements(By.className("momentsOverview__loadMoreBtn___2gLGP")); //$NON-NLS-1$
        }

        // get main rows
        return list.findElements(By.tagName("li")); //$NON-NLS-1$
    }

    /**
     * Creates an {@link Umsatz} from the given WebElement
     *
     * @param listElement
     *            the list element
     * @return the created {@link Umsatz}
     * @throws RemoteException
     *             in case the Umsatz could not be created
     */
    private Umsatz getGiroTransactionDetails(final Pair<Date, WebElement> listElement) throws RemoteException {

        // get all lines from teh transaction WebElement
        final String[] parts = getTransactionPartsFromWebElement(listElement.getValue());

        // get a new transaction
        final HBCIDBService dbService = Settings.getDBService();
        final Umsatz umsatz = dbService.createObject(Umsatz.class, null);

        final Date date = listElement.getKey();

        // set date and valuta to the same as we have only one date
        umsatz.setDatum(date);
        umsatz.setValuta(date);

        // Type of transaction
        umsatz.setArt(parts[0]);

        // Amount in EUR
        final String amount = parts[2].substring(0, parts[2].length() - 1);
        umsatz.setBetrag(Utils.convertStringToDouble(amount));

        // Details
        Utils.loadUmsatz(umsatz, parts[3]);

        // return the transaction
        return umsatz;

    }

    private Umsatz getPreBookedDetails(final Pair<Date, WebElement> listElement) throws RemoteException {

        // get the WebElement
        final WebElement dayElement = listElement.getValue();

        // get the lines from the transaction WebElement
        final String[] parts = getTransactionPartsFromWebElement(dayElement);

        // Details as <p> table.... :-(
        final Map<String, String> detailsMap = new ConcurrentHashMap<>();
        final WebElement detailsTable = dayElement.findElement(By.className("cardsListItem__details___3I1PM")); //$NON-NLS-1$
        final List<WebElement> detailsElements = detailsTable.findElements(By.tagName(WebUtils.HTML_TAG_PARAGRAPH));
        for (int i = 0; i < detailsElements.size(); i = i + 2) {
            Logger.debug("Adding transaction detail: " + detailsElements.get(i).getText() + "=" //$NON-NLS-1$ //$NON-NLS-2$
                    + detailsElements.get(i + 1).getText());
            detailsMap.put(detailsElements.get(i).getText(), detailsElements.get(i + 1).getText());
        }

        // create a new transaction in Hibiscus
        final HBCIDBService dbService = Settings.getDBService();
        final Umsatz umsatz = dbService.createObject(Umsatz.class, null);

        // set date
        umsatz.setDatum(listElement.getKey());

        // Type of transaction
        umsatz.setArt(parts[0]);

        // Amount in EUR
        final String amount = parts[2].substring(0, parts[2].length() - 1);
        umsatz.setBetrag(Utils.convertStringToDouble(amount));

        try {
            // Create simple Date formatter
            final SimpleDateFormat sdf = new SimpleDateFormat("dd.mm.yyyy", Locale.getDefault()); //$NON-NLS-1$
            // set valuta
            final Date valuta = sdf.parse(detailsMap.get(TransactionsPage.DATUM_STRING));
            umsatz.setValuta(valuta);
        } catch (final ParseException e) {
            // Only log
            Logger.error("Error pasring valuta date: " + detailsMap.get(TransactionsPage.DATUM_STRING), e); //$NON-NLS-1$
        }

        // zweck
        Utils.loadUmsatz(umsatz, parts[1]);

        // set details in comment
        final StringBuilder comment = new StringBuilder(130);
        comment.append("Ort: ").append(detailsMap.get("Ort")).append("\nKreditkartennummer: ") //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
                .append(detailsMap.get("Kreditkartennummer:")).append("\nDatum: ") //$NON-NLS-1$ //$NON-NLS-2$
                .append(detailsMap.get(TransactionsPage.DATUM_STRING)).append("\nWhrung: ") //$NON-NLS-1$
                .append(detailsMap.get("Whrung:")).append("\nBetrag in auslndischer Whrung: ") //$NON-NLS-1$ //$NON-NLS-2$
                .append(detailsMap.get("Betrag in auslndischer Whrung:")).append("\nWechselkurs: ") //$NON-NLS-1$ //$NON-NLS-2$
                .append(detailsMap.get("Wechselkurs:")); //$NON-NLS-1$
        umsatz.setKommentar(comment.toString());

        // mark as not booked
        umsatz.setFlags(Umsatz.FLAG_NOTBOOKED);

        // return the transaction
        return umsatz;

    }

    /**
     * Return a paired list of all transactions with the date. The date needs to
     * be calculated as the year information is missing.
     *
     * @param listElements
     *            list with all elements
     * @return a paired list with Date,Transaction (still as WebElement)
     * @throws RemoteException
     *             in case something went wrong with the date calculation
     */
    private List<Pair<Date, WebElement>> getTransactionList(final List<WebElement> listElements)
            throws RemoteException {

        // create a new empty list
        final List<Pair<Date, WebElement>> list = new ArrayList<>();

        // differentiate between giro and credit card
        String className;
        if (this.isCreditCard) {
            className = "cardsListItem__wrapper___25_YU"; //$NON-NLS-1$
        } else {
            className = "accountsListItem__listItemWrapper___1rl7c"; //$NON-NLS-1$
        }

        // set actual year and month for year calculations
        this.year = Calendar.getInstance().get(Calendar.YEAR);
        this.lastTransactionMonth = Calendar.getInstance().get(Calendar.MONTH);

        // Loop over all list elements
        for (final WebElement listElement : listElements) {

            // get the text from the transaction
            final String test = listElement.getText();
            Logger.debug(TransactionsPage.LOG_UMSATZ_ALS_TEXT + test);
            // split into parts (lines)
            final String[] parts = test.split("\n"); //$NON-NLS-1$

            // set date and valuta to the same as we have only one date
            final Date date = calculateDate(parts[0], parts[1]);

            // Get all entries for this date
            Logger.debug("Such Eintrge fr den " + date.toString()); //$NON-NLS-1$

            // get all transactions from this day as it might be severals
            final List<WebElement> dailyList = listElement.findElements(By.className(className));
            for (final WebElement day : dailyList) {
                // add the new pair
                list.add(new MutablePair<>(date, day)); // NOPMD on 14.11.19,
                                                        // 18:49
            }

        }

        // Return the list
        return list;
    }

    /**
     * Returns an array with all lines of the transaction from the given
     * WebElement
     *
     * @param dayElement
     *            the {@link WebElement} with the transaction information
     * @param arrowID
     *            id of the arrow
     * @return an String array with all lines of the transaction
     */
    private String[] getTransactionPartsFromWebElement(final WebElement dayElement) {

        // wait for the overlay to be removed
        SantanderWebSite.waitForLoadingOverlay(this.driver);

        // create a new waiter
        final WebDriverWait wait = new WebDriverWait(this.driver, 10);

        // Scroll, wait and click on list (day) element
        WebUtilsSelenium.scrollTo(this.driver, dayElement);
        wait.until(ExpectedConditions.elementToBeClickable(dayElement));
        dayElement.click();

        // wait for the overlay to be removed and scroll the complete element
        // into view again
        SantanderWebSite.waitForLoadingOverlay(this.driver);
        WebUtilsSelenium.scrollTo(this.driver, dayElement);

        // Log the HMTL of the element
        Logger.trace(String.format("DayElement as XML: \n %s", dayElement.getAttribute("outerHTML"))); //$NON-NLS-1$ //$NON-NLS-2$

        // get the text from the transaction on the web page
        final String test = dayElement.getText();
        Logger.debug(TransactionsPage.LOG_UMSATZ_ALS_TEXT + test);

        // return the the different parts (lines)
        return test.split("\n"); //$NON-NLS-1$
    }

    /**
     * Returns a list with Hibicus transactions {@link Umsatz} created out of
     * the given pair list (Date,WebElement).
     *
     * @param transactionList
     *            th list with the pairs ({@link Date}, {@link WebElement})
     * @param lastTransactionDate
     *            oldest date to give transactions back
     * @return the List with Hibiscus transactions
     * @throws RemoteException
     *             in case some database actions while creating and storing the
     *             transaction informations failed
     */
    private List<Umsatz> getTransactionsFromPairList(final List<Pair<Date, WebElement>> transactionList,
            final Date lastTransactionDate) throws RemoteException {

        final List<Umsatz> transactions = new ArrayList<>();

        // set actual year and month for year calculations
        this.year = Calendar.getInstance().get(Calendar.YEAR);
        this.lastTransactionMonth = Calendar.getInstance().get(Calendar.MONTH);

        // wait a bit
        Utils.waitIgnoreThreadInterrupts(1000);

        // Loop over all list elements
        for (final Pair<Date, WebElement> listElement : transactionList) {

            Umsatz transaction;
            if (this.isCreditCard) {
                // create the transaction from the element
                transaction = getPreBookedDetails(listElement);
            } else {
                // create the transaction from the element
                transaction = getGiroTransactionDetails(listElement);
            }
            // check if search period has been past
            if (transaction.getDatum().before(lastTransactionDate)) {
                break;
            }
            // add transaction to list
            transactions.add(transaction);

        }

        // Return the list
        return transactions;
    }

    /**
     * Open the page for the given account.
     *
     * @throws WebSiteException
     *             in case the page cannot be opened or the account is not found
     */
    private void openPage(final String ktoIdentifier) throws WebSiteException {

        try {
            final String callUrl = this.baseUrl + "/global-position"; //$NON-NLS-1$

            this.driver.get(callUrl);

            // wait for loading is finished
            SantanderWebSite.waitForLoadingOverlay(this.driver);

            final WebDriverWait wait10Sec = new WebDriverWait(this.driver, 120);

            wait10Sec.until(ExpectedConditions
                    .presenceOfElementLocated(By.className("globalPosition__fixedContainer___EM8Do"))); //$NON-NLS-1$ )

            Logger.debug(this.driver.getPageSource());

            Logger.debug(String.format("Search for account id: %S", ktoIdentifier)); //$NON-NLS-1$

            final List<WebElement> spans = this.driver.findElements(By.tagName("span")); //$NON-NLS-1$
            WebElement accountLink = null;
            for (final WebElement webElement : spans) {
                String text = webElement.getText().trim();
                text = text.replaceAll(" ", StringUtils.EMPTY); //$NON-NLS-1$
                Logger.debug("Span: " + text); //$NON-NLS-1$
                if (ktoIdentifier.equalsIgnoreCase(text)) {
                    accountLink = webElement;
                    break;
                }
            }
            if (accountLink == null) {
                throw new WebSiteException(I18N.tr("AccountPage.accountNotFound", ktoIdentifier), //$NON-NLS-1$
                        WebSiteExceptionTypes.KONTO_NOT_FOUND);
            }
            // Wait the overlay
            SantanderWebSite.waitForLoadingOverlay(this.driver);
            accountLink.click();
            wait10Sec.until(ExpectedConditions
                    .visibilityOfElementLocated(By.className("movementsHeader__detailsButton___3I8RQ"))); //$NON-NLS-1$

        } catch (final ElementNotFoundException | ElementNotVisibleException e) {
            throw new WebSiteException(I18N.tr("AccountPage.elementNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.OPERATION_NOT_SUPPORTED, e);

        }

    }

}
