/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.gui.controller;

import java.rmi.RemoteException;

import org.apache.commons.lang.StringUtils;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;

import de.frankmuenster.mahoe.assistant.CreateInfo;
import de.frankmuenster.mahoe.utils.I18N;
import de.willuhn.jameica.gui.AbstractControl;
import de.willuhn.jameica.gui.input.CheckboxInput;
import de.willuhn.jameica.gui.input.IntegerInput;
import de.willuhn.jameica.gui.input.TextInput;
import de.willuhn.jameica.hbci.gui.filter.KontoFilter;
import de.willuhn.jameica.hbci.gui.input.KontoInput;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.server.KontoauszugPdfUtil;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;

/**
 * @author Frank Mnster
 * @param <I>
 *            the concrete CreateInfo
 */
public abstract class AbstractSettingsController<I extends CreateInfo> extends AbstractControl
        implements SettingsController {

    /**
     * Kontoinformationen
     */
    protected final transient I ktoInfo;

    /**
     * (Kreditkarten)-Konto
     */
    protected transient KontoInput account;

    /**
     * Kontotyp
     */
    private transient TextInput ktoType;

    /**
     * Die BLZ des Bank fr den Kontroller
     */
    private final transient String blz;

    /**
     * Kontoauszge speichern
     */
    private transient CheckboxInput savePDF;

    /**
     * Period to search back for account statements. Default is 90 days back.
     */
    protected transient IntegerInput statementsMaxDaysBack;

    /**
     * Should account type be displayed. Default is <i>false</i>
     */
    private boolean displayAccountType;

    /**
     * Should account statement section be displayed. Default is <i>true</i>
     */
    private boolean displayAccountStatement = true;

    /**
     * Standardkonstruktor
     *
     * @param blz
     * @param ktoInfo
     */
    public AbstractSettingsController(final String blz, final I ktoInfo) {

        super(null);
        this.ktoInfo = ktoInfo;
        this.blz = blz;
    }

    /**
     * Liefert ein Kontoauswahlfeld mit allen Konten dieser Bank.
     *
     * @return ein Kontoauswahlfeld mit allen Konten dieser Bank.
     */
    @Override
    public KontoInput getKonto() {

        if (this.account == null) {
            try {
                final KontoFilter ktoFilter = new KontoFilter() {

                    /**
                     * @see de.willuhn.jameica.hbci.gui.filter.KontoFilter#accept(Konto)
                     */
                    @Override
                    public boolean accept(final Konto konto) throws RemoteException {

                        // Muss auf jeden Fall erst mal ein Online-Konto sein.
                        if (!KontoFilter.ONLINE.accept(konto)) {
                            return false;
                        }
                        // Jetzt noch checken, ob die BLZ zur Santander passt
                        final String blz1 = konto.getBLZ();
                        return StringUtils.isNotEmpty(blz1) && blz1.equals(AbstractSettingsController.this.blz);
                    }
                };
                this.account = new KontoInput(null, ktoFilter);
                this.account.addListener(new Listener() {

                    /**
                     * @see Listener#handleEvent(Event)
                     */
                    @Override
                    public void handleEvent(final Event event) {

                        AbstractSettingsController.this.handleAccountChanged(event);
                    }
                });
            } catch (final RemoteException e) {
                final String msg = I18N.tr("SettingsController.createAccountListFailed"); //$NON-NLS-1$
                Logger.error(msg, e);
                final StatusBarMessage msgUser = new StatusBarMessage(msg, StatusBarMessage.TYPE_ERROR);
                Application.getMessagingFactory().sendMessage(msgUser);
            }
            final Event event = new Event();
            event.type = SWT.Selection;
            this.handleAccountChanged(event);
        }
        return this.account;
    }

    /**
     * Liefert ein TextInput mit der Kontoart
     *
     * @return ein TextInput mit der Kontoart
     */
    public TextInput getKontoTyp() {

        if (this.ktoType == null) {
            this.ktoType = new TextInput(StringUtils.EMPTY);
            this.ktoType.setEnabled(false);
            if (this.ktoInfo != null && this.ktoInfo.getAccountType() != null) {
                this.ktoType.setValue(this.ktoInfo.getAccountType().getName());
            }
        }
        return this.ktoType;
    }

    /**
     * Liefert eine Checkbox zur Auswahl, ob die Kreditkartenauszge gespeichert
     * werden sollen.
     *
     * @return Ein CheckBoxInput zur Auswahl, ob die Kreditkartenauszge
     *         gespeichert werden sollen.
     */
    public CheckboxInput getSavePDF() {

        if (this.savePDF == null) {
            Logger.debug("savePDF: " + this.supportsStatements()); //$NON-NLS-1$
            this.savePDF = new CheckboxInput(this.supportsStatements());
            this.savePDF.disable();
        }
        return this.savePDF;
    }

    /**
     * Returns an {@link IntegerInput} to define the max period in the past to
     * load account sheets
     *
     * @return A IntegerInput to define the max period in the past in days
     */
    public IntegerInput getStatementMaxDaysBack() {

        if (this.statementsMaxDaysBack == null) {
            this.statementsMaxDaysBack = new IntegerInput();
            if (this.ktoInfo != null) {
                this.statementsMaxDaysBack.setValue(this.ktoInfo.getStatementsMaxDaysBack());
            }
            this.statementsMaxDaysBack.setComment(StringUtils.EMPTY);
            this.statementsMaxDaysBack.setEnabled(false);
        }
        return this.statementsMaxDaysBack;

    }

    /**
     * @return the displayAccountStatement
     */
    public boolean isDisplayAccountStatement() {

        return this.displayAccountStatement;
    }

    /**
     * @return the displayAccountType
     */
    public boolean isDisplayAccountType() {

        return this.displayAccountType;
    }

    /**
     * Sets if the account statements sections should bve displayed. Default is
     * <i>true</i>.
     *
     * @param displayAccountStatement
     *            the displayAccountStatement to set.
     */
    public void setDisplayAccountStatement(final boolean displayAccountStatement) {

        this.displayAccountStatement = displayAccountStatement;
    }

    /**
     * Sets if the account type should be displayed. Default is <i>false</i>.
     *
     * @param displayAccountType
     *            the displayAccountType to set.
     */
    public void setDisplayAccountType(final boolean displayAccountType) {

        this.displayAccountType = displayAccountType;
    }

    /**
     * Checks if the current account supports statement downloads.
     *
     * @return <i>true</i> in case the account supports statement downloads,
     *         <i>false</i> else
     */
    @Override
    public boolean supportsStatements() {

        boolean isSupporting = false;
        final Konto konto = (Konto) this.getKonto().getValue();
        if (konto != null) {
            isSupporting = KontoauszugPdfUtil.supported(konto);
        }
        return isSupporting;
    }

    /**
     * Behandelt einen Kontowechsel
     *
     * @param event
     */
    protected abstract void handleAccountChanged(final Event event);
}
