/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.utils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Duration;
import java.util.Arrays;
import java.util.HashSet;

import org.openqa.selenium.JavascriptExecutor;
import org.openqa.selenium.TimeoutException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.FluentWait;
import org.openqa.selenium.support.ui.WebDriverWait;

import de.frankmuenster.mahoe.banks.santander.web.SantanderWebSite;
import de.frankmuenster.mahoe.web.AbstractSeleniumWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Utility class for web methods
 *
 * @author Frank Mnster
 */
public final class WebUtilsSelenium {

    /**
     * The maximum number of new files during a download in the download folder.
     */
    private static final int MAX_NUMBER_NEW_FILES = 1;

    /**
     * Read the account statement PDF into a byte array.
     *
     * @param driver
     *            the {@link WebDriver} to use
     * @param linkElement
     *            the link on the page to the PDF
     * @param expectedFileName
     * @return the PDF as byte array
     */
    public static byte[] getSeleniumPdfAsByteArray(final WebDriver driver, final WebElement linkElement)
            throws WebSiteException {

        byte[] data = null;

        Logger.debug("Downloading file with unknown name"); //$NON-NLS-1$

        try {

            final File dir = AbstractSeleniumWebSite.getDownloadsDirectory().toFile();

            // get the actual files in the download folder
            final HashSet<File> existingFileList = new HashSet<>(Arrays.asList(dir.listFiles()));
            // how many have we
            final int fileCount = existingFileList.size();

            // scroll into view
            WebUtilsSelenium.scrollTo(driver, linkElement);

            // wait for any overlay
            SantanderWebSite.waitForLoadingOverlay(driver);

            final WebDriverWait wait1 = new WebDriverWait(driver, 10);
            wait1.until(ExpectedConditions.elementToBeClickable(linkElement));

            // trigger the download
            linkElement.click();
            // wait until we have more files in the download folder
            final FluentWait<WebDriver> wait = new FluentWait<>(driver).withTimeout(Duration.ofSeconds(60))
                    .pollingEvery(Duration.ofMillis(200));

            // For the file system to be ready
            Utils.waitIgnoreThreadInterrupts(1000);

            Logger.debug("Files in download dir: " + dir.list().length); //$NON-NLS-1$
            wait.until((final WebDriver wd) -> dir.list().length > fileCount);

            // get new file sin folder
            final HashSet<File> newFileList = new HashSet<>(Arrays.asList(dir.listFiles()));
            // remove the existing once
            newFileList.removeAll(existingFileList);
            // check if one 1 new file
            if (newFileList.size() == WebUtilsSelenium.MAX_NUMBER_NEW_FILES) {
                // get the file
                final File downloadedFile = newFileList.iterator().next();
                Logger.info(I18N.tr("WebUtils.downloadedFileName") + downloadedFile.getAbsolutePath()); //$NON-NLS-1$
                // read file in memory
                data = Files.readAllBytes(downloadedFile.toPath());
                // delete file again
                Files.delete(downloadedFile.toPath());
            } else {
                Logger.error(I18N.tr("WebUtils.moreThanOneNewFileInDownloadFolderMsg")); //$NON-NLS-1$
                throw new WebSiteException(I18N.tr("WebUtils.moreThanOneNewFileInDownloadFolderExp"), //$NON-NLS-1$
                        WebSiteExceptionTypes.TRANSFER_FAILED);
            }

        } catch (final IOException | TimeoutException | ApplicationException e) {
            Logger.error(I18N.tr("WebUtils.errorDownloadAccountSheet1"), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("WebUtils.errorDownloadAccountSheet1"), //$NON-NLS-1$
                    WebSiteExceptionTypes.TRANSFER_FAILED, e);
        }

        return data;

    }

    /**
     * Read the account statement PDF into a byte array.
     *
     * @param driver
     *            the {@link WebDriver} to use
     * @param linkElement
     *            the link on the page to the PDF
     * @param expectedFileName
     * @return the PDF as byte array
     */
    public static byte[] getSeleniumPdfAsByteArray(final WebDriver driver, final WebElement linkElement,
            final String expectedFileName) {

        byte[] data = null;

        Logger.debug("Downloading file: " + expectedFileName); //$NON-NLS-1$

        linkElement.click();
        try {

            final FluentWait<WebDriver> wait = new FluentWait<>(driver).withTimeout(Duration.ofSeconds(60))
                    .pollingEvery(Duration.ofMillis(200));
            final File fileToCheck = AbstractSeleniumWebSite.getDownloadsDirectory().resolve(expectedFileName).toFile();

            final Path dir = AbstractSeleniumWebSite.getDownloadsDirectory();

            dir.getNameCount();

            // Wait until download is complete
            wait.until((final WebDriver wd) -> fileToCheck.exists());
            // read file in memory
            data = Files.readAllBytes(fileToCheck.toPath());
            // delete file again
            Files.delete(fileToCheck.toPath());

        } catch (final ApplicationException | IOException | TimeoutException e) {
            Logger.error(I18N.tr("WebUtils.errorDownloadAccountSheet", expectedFileName), e); //$NON-NLS-1$
        }

        return data;

    }

    /**
     * Scroll to an element so it is visible
     */
    public static void scrollTo(final WebDriver driver, final WebElement element) {

        // Scroll and click on list element
        final JavascriptExecutor executor = (JavascriptExecutor) driver;
        executor.executeScript("arguments[0].scrollIntoView(false);", //$NON-NLS-1$
                element);

    }

    /**
     * private class constructor
     */
    private WebUtilsSelenium() {

        // intentionally left blank
    }

}
